"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Debug = require("debug");
const debug = Debug("blockchain:handler");
const utils_1 = require("./utils");
class BlockHandler {
    constructor(blockchain, db, difficulty) {
        this.blockchain = blockchain;
        this.db = db;
        this.proofRegex = this.buildRegex(difficulty || 4);
        // TODO: BlockHandler <-> Blockchain relationship could be refactored
    }
    /**
     * returns the hash for a block
     * @param block
     */
    hash(block) {
        if (block === null) {
            throw new Error("Cannot hash block that is null.");
        }
        const blockString = JSON.stringify(block);
        return utils_1.getSHA256Hash(blockString);
    }
    /**
     * creates a new block on the blockchain
     * @param proof
     * @param previousHash
     */
    async newBlock(proof, previousHash) {
        const block = {
            index: await this.blockchain.getNextBlockIndex(),
            previousHash: previousHash || this.hash(await this.lastBlock()),
            proof,
            timestamp: Date.now(),
            transactions: this.blockchain.getCurrentTransactions(),
        };
        debug("creating new block");
        this.blockchain.clearTransactions();
        await this.blockchain.storeBlock(block);
        return block;
    }
    /**
     * gets the last block of the chain
     */
    async lastBlock() {
        return await this.blockchain.getLastBlock();
    }
    /**
     * simple proof of work algorithm
     * find a number p such that hash(pq) contains
     * certain amount of leading zeros, where previous p
     * is q
     * @param lastProof previous proof
     */
    proofOfWork(lastProof) {
        debug("searching for proof for", lastProof);
        const startT = Date.now();
        let proof = 0;
        while (!this.isProofValid(lastProof, proof)) {
            proof++;
        }
        const duration = Date.now() - startT;
        debug("found proof", proof, "for previous", lastProof, "in", duration, "ms");
        return proof;
    }
    /**
     * validates the proof
     * if the hash of the last proof and new proof
     * starts with 4 zeros
     * @param lastProof
     * @param proof
     */
    isProofValid(lastProof, proof) {
        const guess = `${lastProof}:${proof}`;
        const guessHash = utils_1.getSHA256Hash(guess);
        return this.proofRegex.test(guessHash);
    }
    /**
     * validates a block and its transactions
     * @param block
     */
    isBlockValid(block) {
        if (!block || !block.previousHash || !block.proof ||
            !block.transactions || !block.transactions.length) {
            return false;
        }
        let invalidTransactionFound = false;
        block.transactions.forEach((transaction) => {
            if (!this.blockchain.signature.verifyTransactionSignature(transaction)) {
                invalidTransactionFound = true;
            }
        });
        if (invalidTransactionFound) {
            debug("block contains invalid transaction");
            return false;
        }
        return true;
    }
    /**
     * checks if the current chain is valid
     */
    async isChainValid(chain) {
        let index = 1;
        while (index < chain.length) {
            const previousBlock = chain[index - 1];
            const block = chain[index];
            // check that all hashes of blocks are correct
            if (block.previousHash !== this.hash(previousBlock)) {
                return false;
            }
            // check that the proof of work of the block is correct
            if (!this.isProofValid(previousBlock.proof, block.proof)) {
                return false;
            }
            index++;
        }
        return true;
    }
    /**
     * generates a regex to validate
     * the proof of work (leading zeros)
     * @param difficulty
     */
    buildRegex(difficulty) {
        const regHashStart = [];
        for (let i = 0; i < difficulty; i++) {
            regHashStart.push("0");
        }
        return new RegExp(`^${regHashStart.join("")}`);
    }
}
exports.default = BlockHandler;
//# sourceMappingURL=BlockHandler.js.map