"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Debug = require("debug");
const debug = Debug("blockchain:blockchain");
const BlockHandler_1 = require("./BlockHandler");
const Signature_1 = require("./crypto/Signature");
const Address_1 = require("./crypto/Address");
const Database_1 = require("./db/Database");
const Peer_1 = require("./Peer");
class Blockchain {
    constructor(config) {
        this.config = config;
        const { difficulty, nodeAddress } = this.config.blockchain;
        this.db = new Database_1.default(config);
        this.blockHandler = new BlockHandler_1.default(this, this.db, difficulty);
        this.signature = new Signature_1.default(config);
        this.address = new Address_1.default(config);
        // if no address is configured, generate a new one
        this.nodeAddress = nodeAddress || this.address.createAddress();
        this.currentTransactions = [];
        this.conflictResolvationRunning = false;
        this.p2p = new Peer_1.default(config);
    }
    removeTransactionForSignature(signature) {
        let i = 0;
        for (const transaction of this.currentTransactions) {
            if (transaction.signature === signature) {
                this.currentTransactions.splice(i, 1);
                return true;
            }
            i++;
        }
        return false;
    }
    /**
     * adds a new node to the node list (is proxy function)
     * @param address address of the node e.g. http://localhost:1337
     */
    registerNode(address) {
        this.p2p.registerNode(address);
    }
    /**
     * removes a node from the node list (is proxy function)
     * @param address
     */
    removeNode(address) {
        this.p2p.removeNode(address);
    }
    /**
     * returns false if this node is currently busy
     */
    doesAcceptOperations() {
        return !this.conflictResolvationRunning;
    }
    /**
     * register this node on another using the advertised hostname (is proxy function)
     * @param host
     */
    async registerSelfAtOtherNode(host) {
        return await this.p2p.registerSelfAtOtherNode(host);
    }
    /**
     * resolve conflicts faster by only fetching the chain sizes
     * from the nodes
     */
    async resolveConflicts() {
        debug("resolving conflicts fast.");
        if (this.conflictResolvationRunning) {
            debug("conflict resolvation still running.");
            return false;
        }
        this.conflictResolvationRunning = true;
        const resolveRecursive = async () => {
            if (!this.getNodesCount()) {
                debug("no more nodes left");
                return false;
            }
            const { host } = await this.findLeadingNode();
            if (host === null) {
                debug("current node is leading node.");
                return false;
            }
            const chain = await this.p2p.fetchChainFromNode(host);
            if (!chain || !chain.length) {
                // if fetch fails, remove node and try again
                debug("fetch failed, removing node");
                this.removeNode(host);
                return resolveRecursive();
            }
            const oldLength = await this.getLength();
            // chain is too large to handle, store it in db first
            // (TODO: this should be rebuild to streaming chunks for future scaling)
            await this.db.replaceFullChain(chain);
            const newLength = await this.getLength();
            debug("replacing chain with", newLength, "from", host);
            if (newLength < oldLength) {
                debug("fetched chain is smaller than expected, removing node");
                this.removeNode(host);
                return resolveRecursive();
            }
            // TODO: fetching the whole chain will not scale
            // fetching headers will also not help, as the transactions
            // are needed to create the hash to compare to the following block
            const blockchain = await this.db.getFullDBAsBlocks();
            if (!this.blockHandler.isChainValid(blockchain)) {
                debug("fetched chain is invalid, removing node");
                this.removeNode(host);
                return resolveRecursive();
            }
            debug("fetched and exchanged chain is valid, synced with", host);
            return true;
        };
        return resolveRecursive().then((result) => {
            this.conflictResolvationRunning = false;
            return result;
        }).catch((error) => {
            this.conflictResolvationRunning = false;
            debug("conflict resolvation failed with error", error.message);
            return false;
        });
    }
    /**
     * searches for the leading node amonst all known nodes
     * using their chains length
     */
    async findLeadingNode() {
        debug("searching for leading node");
        const fetchPromises = [];
        this.p2p.nodes.forEach((host) => {
            fetchPromises.push(this.p2p.fetchChainLengthFromNode(host).then((length) => {
                return {
                    host,
                    length,
                };
            }));
        });
        return Promise.all(fetchPromises).then(async (results) => {
            let leadingHost = null;
            let maxLength = await this.getLength();
            results.forEach(({ host, length }) => {
                if (length > maxLength) {
                    leadingHost = host;
                    maxLength = length;
                }
            });
            return {
                host: leadingHost,
                length: maxLength,
            };
        });
    }
    /**
     * inform all other known nodes about the transaction
     * @param transaction
     * @param signature
     */
    async publishTransactionToKnownNodes(transaction) {
        const publishPromises = [];
        this.p2p.nodes.forEach((host) => {
            publishPromises.push(this.p2p.publishTransaction(transaction, host));
        });
        return Promise.all(publishPromises);
    }
    /**
     * inform all other nown nodes about the block
     * @param block
     */
    async publishBlockToKnownNodes(block) {
        const publishPromises = [];
        this.p2p.nodes.forEach((host) => {
            publishPromises.push(this.p2p.publishBlock(block, host));
        });
        return Promise.all(publishPromises);
    }
    applyTransactionInformation(newTransaction) {
        let found = false;
        this.currentTransactions.forEach((transaction) => {
            if (newTransaction.sender === transaction.sender &&
                newTransaction.recipient === transaction.recipient) {
                found = true;
            }
        });
        if (found) {
            debug("transaction already exists, wont add.");
            return false;
        }
        debug("transaction does not exist, adding");
        this.addTransaction(newTransaction);
        return true;
    }
    async applyBlockInformation(block) {
        if (!this.blockHandler.isBlockValid(block)) {
            debug("block from other node is invalid, wont add", block.index);
            return false;
        }
        if (await this.db.checkIfBlockExists(block)) {
            debug("block already exists", block.index);
            return false;
        }
        // remove all transactions of this block from the current transactions
        // of the next block of this node, in case there are any
        block.transactions.forEach((transaction) => {
            this.removeTransactionForSignature(transaction.signature);
        });
        await this.db.storeBlock(block);
        debug("added block from other node", block.index);
        return true;
    }
    getNodesAsList() {
        const list = [];
        this.p2p.nodes.forEach((host) => {
            list.push(host);
        });
        return list;
    }
    getNodesCount() {
        return this.p2p.nodes.size;
    }
    getBlockHandler() {
        return this.blockHandler;
    }
    getDatabase() {
        return this.db;
    }
    /**
     * db-proxy
     */
    async getLength() {
        return this.db.getChainLength();
    }
    async getNextBlockIndex() {
        return (await this.getLength()) + 1;
    }
    getCurrentTransactions() {
        return this.currentTransactions;
    }
    clearTransactions() {
        this.currentTransactions = [];
    }
    /**
     * db-proxy
     */
    async getLastBlock() {
        return this.db.getLastBlock();
    }
    /**
     * db-proxy
     */
    streamChainFromDB() {
        return this.db.getWholeChain();
    }
    /**
     * db proxy
     * @param block
     */
    async storeBlock(block) {
        await this.db.storeBlock(block);
    }
    /**
     * db proxy
     * @param address
     */
    async getAddressBalance(address) {
        return await this.db.getBalanceOfAddress(address);
    }
    /**
     * db proxy
     * @param address
     */
    async getAddressTransactions(address) {
        return await this.db.getTransactionsOfAddress(address);
    }
    /**
     * creates a new block (stores it in database) and publishes it
     */
    async addBlock(proof, previousHash, publish = true) {
        const block = await this.blockHandler.newBlock(proof, previousHash);
        if (publish) {
            this.publishBlockToKnownNodes(block).catch(() => {
                // empty
            });
        }
        return block;
    }
    /**
     * adds a new transaction and publishes it
     * to the other known nodes
     * @param transaction
     * @param publish
     */
    async addTransaction(transaction, publish = true) {
        this.currentTransactions.push(transaction);
        if (publish) {
            this.publishTransactionToKnownNodes(transaction).catch(() => {
                // empty
            });
        }
        return await this.getNextBlockIndex();
    }
    async init() {
        await this.db.init();
        const lastBlock = await this.db.getLastBlock();
        if (!lastBlock) {
            this.addTransaction({
                amount: 0,
                sender: this.config.blockchain.mintAddress,
                recipient: this.config.blockchain.mintAddress,
                timestamp: Date.now(),
                payload: "",
                signature: "",
            }, false);
            await this.blockHandler.newBlock(100, "1");
            debug("no block present, created genisis block");
        }
        else {
            debug("blocks present, wont create genisis block");
        }
        // in case that other nodes are configured, we can get updated
        await this.resolveConflicts();
    }
    close() {
        this.db.close();
    }
}
exports.default = Blockchain;
//# sourceMappingURL=Blockchain.js.map